#
# START COPYRIGHT-NOTICE: 2002-2003
#
# Copyright 2002-2003 VERITAS Software Corporation.
# All rights reserved.
#
# VERITAS, VERITAS SOFTWARE, the VERITAS logo and all other VERITAS
# product names and slogans are trademarks or registered trademarks
# of VERITAS Software Corporation in the USA and/or other countries.
# Other product names and/or slogans mentioned herein may be trade-
# marks or registered trademarks of their respective companies.
#
# END COPYRIGHT-NOTICE.
#

package SRDFAgent;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use ag_i18n_inc;
my $vcs_home=$ENV{"VCS_HOME"} || "/opt/VRTSvcs";

@EXPORT = qw(
);
$VERSION = '1.00';

require 5.000;

%default = {
	resname => '',
	name	=> '',
	home	=> '',
	mode	=> '',
	state 	=> '',
	role 	=> '',
	devs	=> 0,
	invalid	=> 0,
	devfotime => 0
};

sub new {
	my($proto, $resname, $name, $symhome, $mode, $devfotime) = @_;
	my $class = ref($proto) || $proto;
	my $self = { %default };
	$self->{'resname'} = $resname;
	$self->{'name'} = $name;
	$self->{'home'} = $symhome;
	$self->{'mode'} = $mode;
	$self->{'start_time'} = $^T;
	if ($self->{'mode'} eq ''){
		$self->{'mode'} = "MODE_NOT_SPEC";
	}
	$self->{'devfotime'} = $devfotime;
	bless($self, $class);
	return $self;
}

sub DESTROY {}

sub get_state {
	my $self = shift;
	my $rw = 0;
	my $wd = 0;
	my $cmd = $self->{'home'} . "/bin/symrdf";
	$self->{'state'} = "";
	my @query = split('\n', `$cmd -g $self->{'name'} query`);
	foreach my $q (@query) {
		if ($q =~ /Source/) {
			my @roleinfo = split('\s+', $q);
			if ($roleinfo[1] eq "Source") {
				$self->{'role'} = "r1";
			} else {
				$self->{'role'} = "r2";
			}
			next;
		}
		if (!($q =~/[SAC]\./)) {
			next;
		}
		my @stateinfo = split('\s+', $q);
		if ($stateinfo[2] eq "RW") {
			if ($self->{'state'} eq "WD") {
				$self->{'state'} = "Invalid";
				return;
			} else {
				$self->{'state'} = "RW";
			}
		}
		if ($stateinfo[2] eq "WD") {
			if ($self->{'state'} eq "RW") {
				$self->{'state'} = "Invalid";
				return;
			} else {
				$self->{'state'} = "WD";
			}
		}
		++$self->{'devs'};
		$self->{'invalid'} = $self->{'invalid'} + $stateinfo[3];
	}
	if (!$self->{'state'}) {
		$self->{'state'} = "Invalid";
	}
}

sub get_mode {
	my $self = shift;
	my $retVal = 0; # 0 stands for Synchronous
	my $rdf_mode = '';

	my $cmd = $self->{'home'} . "/bin/symdg show " . $self->{'name'};
	my @query = split('\n', `$cmd`);
	foreach my $q (@query) {
		if ($q =~ /RDF Mode/) {
			my @modeinfo = split('\s+', $q);
			$rdf_mode = $modeinfo[4];
			last;
		}
	}


	# Check the specified mode with the discovered mode. The two should match. 
	# If not, then give an error and exit.
	#
	# Partitioned state is a very special case. 
	#
	# In partitioned state, for asynchronous mode of replication, the mode on
	# the R2 side changes to Synchronous (actually it should drop to adaptive
	# copy, but it drops to the default of "S" i.e. synchronous.)
	#
	# To consider the above case, we do not signal an error if -
	#
	# - the devices are in a partitioned state.
	# - The mode specified in the configuration is 'Asynchronous'
	# - The discovered mode is 'Synchronous'
	#
	# However, in a partitioned state, we will signal an error if -
	#
	# 1. The configured state is 'Synchronous' and the discovered state 
	#    is 'Asynchronous'
	#
	# +-------------------------------------------------------------+
	# |   Configured State  |  Discovered State | Result            |
	# |---------------------+-------------------+-------------------|
	# |                     |                   |                   |
	# |   Synchronous       | Synchronous       | Valid             |
	# |---------------------+-------------------+-------------------|
	# |   Synchronous       | Asynchronous      | Error             |
	# |---------------------+-------------------+-------------------|
	# |   Asynchronous      | Synchronous       | Valid if          |
	# |                     |                   | partitioned state |
	# |                     |                   |                   |
	# |                     |                   | Error if not      |
	# |                     |                   | partitioned state |
	# |                     |                   |                   |
	# |---------------------+-------------------+-------------------|
	# |   Asynchronous      | Asynchronous      | Valid             |
	# +-------------------------------------------------------------+
	#
	if ($self->{'mode'} ne "MODE_NOT_SPEC") {

		if (($self->{'mode'} eq $rdf_mode) && ($self->{'mode'} eq "Asynchronous")) {
			$retVal = 1;
		}

		if (($self->{'mode'} ne $rdf_mode) && ($self->{'mode'} eq "Synchronous")) {
			# Error - Configured is Sync and discovered is Async.
			VCSAG_LOG_MSG("E", "Inconsistent mode of replication detected. Configured is $self->{'mode'}, detected is $rdf_mode", 24, "$self->{'mode'}", "$rdf_mode");
			exit (0);
		}
		if (($self->{'mode'} ne $rdf_mode) && ($self->{'mode'} eq "Asynchronous")) {
			# Configured is Asynch and discovered is sync.
			my $partitioned = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -partitioned verify > /dev/null 2>&1");
			if ($partitioned != 0) {
				# Error if not partitioned.
				VCSAG_LOG_MSG("E", "Inconsistent mode of replication detected. Devices in group $self->{'name'} are not in partitioned state, mode of replication configured is $self->{'mode'}, detected is $rdf_mode", 25, "$self->{'name'}", "$self->{'mode'}", "$rdf_mode");
				exit (0);
			}
			VCSAG_LOG_MSG("N", "Devices in group $self->{'grpname'} are in partitioned state, mode of replication configured is $self->{'mode'}, detected is $rdf_mode", 26, "$self->{'name'}", "$self->{'mode'}", "$rdf_mode");
			$retVal = 1;
		}
	} else {
		if ($rdf_mode eq "Asynchronous") {
			$retVal = 1;
		}
	}
	
	return $retVal;
}

sub get_consistency {
	my $self = shift;
	my $ret = 0;

	my $cmd = $self->{'home'} . "/bin/symdg show " . $self->{'name'};
	my @query = split('\n', `$cmd`);
	foreach my $q (@query) {
		if (($q =~ /Session Consistency State/) && ($q =~ /Enabled/)) {
			return 1;
		}
	}
	return 0;
}


sub failover_required {
	my ($self) = shift;
	if ($self->{'state'} eq "Invalid") {
		return 2;
	}

        if (($self->{'state'} eq "RW") && ($self->{'role'} eq "r2")) {
                my $ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -split verify 2>/dev/null");
                if ($ret == 0) {
                        return 3;
                }
        }
	if ($self->{'state'} eq "WD") {
		return 1;
	}
	return 0;
}

sub failover_time {
	my ($self) = @_;
	my ($oto) = split('\n', `$vcs_home/bin/hatype -value SRDF OnlineTimeout`);
	my $cur_time = time();
	my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
	my $failover_time = $self->{'devs'} * $self->{'devfotime'};
	my $count = int(($remaining_time - ($failover_time + 10)) / 5);
	if ($count) {
		$count = $count - 1;
	}
	return $count;
}

sub get_count {
	my ($self) = @_;
	my ($oto) = split('\n', `$vcs_home/bin/hatype -value SRDF OnlineTimeout`);
	my $cur_time = time();
	my $remaining_time = ($oto - ($cur_time - $self->{'start_time'}));
	my $count = int($remaining_time / 5) - 1;
	if ($count <= 0) {
		return 0;
	}
	return $count;
}

sub wait_for_sync {
	my ($self) = @_;
	my $count = $self->get_count();
	if ($self->{'role'} eq "r1") {
		my $ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -updateinprog verify 2>/dev/null");
		$ret = $ret / 256;
		if ($ret != 36) {
			VCSAG_LOG_MSG("N", "failback of device group $self->{'name'} will wait until all updates in progress complete or entry point times out", 19, "$self->{'name'}");
			my $ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -updated verify 2>/dev/null");
			if ($ret) {
				return 1;
			}
		}
	} else {
		my $ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -syncinprog verify 2>/dev/null");
		$ret = $ret / 256;
		if ($ret != 28) {
			VCSAG_LOG_MSG("N", "failover of device group $self->{'name'} will wait until all synchronization in progress complete or entry point times out", 20, "$self->{'name'}");
			if ($self->get_mode()) {
				$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -consistent verify 2> /dev/null");
				if ($ret) {
					$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -synchronized verify 2> /dev/null");
				}
			} else {
				$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -synchronized verify 2> /dev/null");
			}

			if ($ret) {
				return 1;
			}
		}
	}
	return 0;
}

sub takeover {
	my ($self, $autotakeover) = @_;
	my $count;
	my $failovercount;

	my $ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -partitioned verify > /dev/null 2>&1");
	if ($ret == 0) {
		if ($autotakeover) {
			# Synchronous mode and invalid tracks found.
			if (!($self->get_mode()) && ($self->{'invalid'})) {
				VCSAG_LOG_MSG("E", "devices in group $self->{'name'} are in Partitioned state and invalid tracks exist; takeover cannot occur and administrative intervention to restore consistent data is required", 11, "$self->{'name'}");
				exit(0);
			} else {
				# Asynchronous mode and invalid tracks found.
				# OR
				# Synchronous mode and no invalid tracks found.
				$failovercount = $self->failover_time();
				if ($failovercount <= 0) {
					VCSAG_LOG_MSG("W", "not enough time to issue failover command; administrative intervention or automatic restart required", 27);
					exit(0);
				}

				VCSAG_LOG_MSG("N", "devices in group $self->{'name'} are in Partitioned state, and autotakeover is enabled; failover of devices will proceed", 20, "$self->{'name'}");
				$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -c $failovercount -i 5 failover"); 
				if ($ret) {
					VCSAG_LOG_MSG("E", "failover of Partitioned devices in group $self->{'name'} failed; administrative intervention is required", 12, "$self->{'name'}");
					exit(0);
				}
				VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 21, "$self->{'name'}");
				$self->create_lockfile();
				$self->cvm_import();
				exit(0);
			}
		} else {
			VCSAG_LOG_MSG("E", "devices in group $self->{'name'} are in Partitioned state but autotakeover is not enabled; administrative intervention is required", 13, "$self->{'name'}");
			exit(0);
		}
	}
}

sub failback {
	my ($self) = @_;
	my $count;
	my $failbackcount;

	$count = $self->get_count();
	my $ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -failedover verify > /dev/null 2>&1");
	if ($ret) {
		$count = $self->get_count();
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -updated verify > /dev/null 2>&1");
		if ($ret) {
			VCSAG_LOG_MSG("E", "all devices in group $self->{'name'} are not in Failed Over or Updated state in prepration for failback; administrative intervention is required", 14, "$self->{'name'}");
			exit(0);
		}
	}
	$failbackcount = $self->failover_time();
	if ($failbackcount <= 0) {
		VCSAG_LOG_MSG("W", "not enough time to issue failback command; administrative intervention or automatic restart required", 28);
		exit(0);
	}

	VCSAG_LOG_MSG("N", "issuing failback of device group $self->{'name'}", 15, "$self->{'name'}");
	$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -i 5 -c $failbackcount failback");
	$self->get_state();
	if ($self->{'state'} eq "RW") {
		VCSAG_LOG_MSG("N", "failback of devices in group $self->{'name'} succeeded", 22, "$self->{'name'}");
		$self->create_lockfile();
		$self->cvm_import();
		exit(0);
	} else {
		$ret = $ret / 256;
		VCSAG_LOG_MSG("E", "failback of device group $self->{'name'} failed; error code is $ret", 16, "$self->{'name'}");
		exit(0);
	}
}

sub failover {
	my ($self) = @_;
	my $count;
	my $failovercount;

	my $SwapFile = $vcs_home . "/bin/SRDF/.VCS_SRDF_" . $self->{'resname'} . "_Dynamic_SwapOK.DO-NOT-REMOVE";

	my $ret = 0;
	$count = $self->get_count();
	if ($self->get_mode()) {
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -consistent verify > /dev/null 2>&1");
		if ($ret) {
			$count = $self->get_count();
			$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -synchronized verify > /dev/null 2>&1");
		}
	} else {
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -c $count -i 5 -synchronized verify > /dev/null 2>&1");
	}
	if ($ret) {
		VCSAG_LOG_MSG("W", "all devices in group $self->{'name'} are not in synchronized or consistent state in prepration for failover; administrative intervention is required", 17, "$self->{'name'}");
		exit(0);
	}

	$failovercount = $self->failover_time();
	if ($failovercount <= 0) {
		VCSAG_LOG_MSG("W", "not enough time to issue failover command; administrative intervention or automatic restart required", 27);
		exit(0);
	}

	if ( ! -f "$SwapFile" ) {
		# No Swap personality and normal failover.
		VCSAG_LOG_MSG("N", "issuing failover of device group $self->{'name'}", 18, "$self->{'name'}");
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -i 5 -c $failovercount failover");
		sleep(3);
		$self->get_state();
		if ($self->{'state'} eq "RW") {
			VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
			$self->create_lockfile();
			$self->cvm_import();
			exit(0);
		} else {
			$ret = $ret / 256;
			VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 19, "$self->{'name'}", "$ret");
			exit(0);
		}
	}

	# For swap devices, if mode is sync.
	if ( !$self->get_mode ()) {
		VCSAG_LOG_MSG("N", "issuing failover/role swap of device group $self->{'name'}", 18, "$self->{'name'}");
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -i 5 -c $failovercount failover -establish");
		sleep(3);
		$self->get_state();
	} else {
		VCSAG_LOG_MSG("N", "issuing -force failover of device group $self->{'name'}", 18, "$self->{'name'}");
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -force -i 5 -c $failovercount failover");
		if ($ret) {
			VCSAG_LOG_MSG("E", "Failover of the device group $self->{'name'} failed; error code is $ret", 19, "$self->{'name'}", "$ret");
			exit (0);
		}

		$count = $self->get_count();
		my $disabled_consistency = 0;
		if ($self->get_consistency()) {
			$disabled_consistency = 1;
			VCSAG_LOG_MSG("N", "disabling consistency for device group $self->{'name'}", 18, "$self->{'name'}");
			$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -i 5 -c $count -noprompt disable");
			if ($ret) {
				VCSAG_LOG_MSG("E", "Disabling the consistency for the device group $self->{'name'} failed; error code is $ret", 19, "$self->{'name'}", "$ret");
				exit (0);
			}
		}

		$count = $self->get_count();
		VCSAG_LOG_MSG("N", "issuing role swap of device group $self->{'name'}", 18, "$self->{'name'}");
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -i 5 -c $count swap");
		if ($ret) {
			VCSAG_LOG_MSG("E", "Role-swap of the device group $self->{'name'} failed; error code is $ret", 19, "$self->{'name'}", "$ret");
			exit (0);
		}
		
		$count = $self->get_count();
		VCSAG_LOG_MSG("N", "issuing establish/resume of device group $self->{'name'}", 18, "$self->{'name'}");
		$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -i 5 -c $count establish");
		if ($ret) {
			VCSAG_LOG_MSG("E", "Establish=link for the device group $self->{'name'} failed; error code is $ret", 19, "$self->{'name'}", "$ret");
			exit (0);
		}
		
		$count = $self->get_count();
		if ($disabled_consistency) {
			VCSAG_LOG_MSG("N", "re-enable consistency for device group $self->{'name'}", 18, "$self->{'name'}");
			$ret = system("$self->{'home'}/bin/symrdf -g $self->{'name'} -noprompt -i 5 -c $count enable");
			if ($ret) {
				VCSAG_LOG_MSG("N", "re-enabling consistency on device group $self->{'name'} failed. Administrative intervention required.", 18, "$self->{'name'}");
				exit(0);
			}
		}
		sleep(3);
		$self->get_state();
	}
	if ($self->{'state'} eq "RW") {
		VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
		$self->create_lockfile();
		$self->cvm_import();
		exit(0);
	} else {
		$ret = $ret / 256;
		VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 19, "$self->{'name'}");
		exit(0);
	}
}
